from pyoperators.utils.mpi import MPI

# force gfortran's read statement to always use the dot sign as fraction
# separator (PR47007)
import locale
locale.setlocale(locale.LC_NUMERIC, 'POSIX')
del locale

from . import _flib
from . import geometry
from .operators import *
from .datatypes import *
from .packedtables import *
from .instruments import *
#from .mpiutils import *
from .acquisitions import *
#from .processing import *
from .quantities import *
from .datautils import *
from .sparse import *
from .wcsutils import *

__version__ = '1.1.dev53'
